
#include <stdio.h>
#include <math.h>

#define BUFSIZE   128
#define MAX_VERTS 255
#define MAXPANELS 40

typedef struct {
    float x, y, z;
} point;

typedef struct {
    point pt[3];		/* Vertices of triangle */
    float area;			/* Unused; might be used for adaptive subdivision */
} triangle;

typedef struct {
    int       npoly;    /* # of triangles in object */
    triangle *poly;		/* Triangles */
} object;

/* Six equidistant points lying on the unit sphere */
#define XPLUS {  1,  0,  0 }    /* X */
#define XMIN  { -1,  0,  0 }    /* -X */
#define YPLUS {  0,  1,  0 }    /* Y */
#define YMIN  {  0, -1,  0 }    /* -Y */
#define ZPLUS {  0,  0,  1 }    /* Z */
#define ZMIN  {  0,  0, -1 }    /* -Z */

/* Vertices of a unit octahedron */
triangle        octahedron[] = {
  {{XPLUS, ZPLUS, YPLUS}, 0.0},
  {{YPLUS, ZPLUS, XMIN}, 0.0},
  {{XMIN, ZPLUS, YMIN}, 0.0},
  {{YMIN, ZPLUS, XPLUS}, 0.0},
  {{XPLUS, YPLUS, ZMIN}, 0.0},
  {{YPLUS, XMIN, ZMIN}, 0.0},
  {{XMIN, YMIN, ZMIN}, 0.0},
  {{YMIN, XPLUS, ZMIN}, 0.0}
};

/* A unit octahedron */
object oct = { sizeof(octahedron) / sizeof(octahedron[0]), &octahedron[0] };

static int curindex = -1;
int line = 0;
int made_sphere = 0;
int sphereprecision = 3;
int cylconeprecision = 8;
int reverse_normals = 0;
int ignore_normals = 0;
int has_alpha = 0;

char *progname, *filename;
char  buf[128];


void
warn(s)
char *s;
{
    fprintf(stderr, "%s [%s]: line %d: %s\n", progname, filename, line, s);
}


void
fatal(s)
char *s;
{
    fprintf(stderr, "%s [%s]: line %d: %s\n", progname, filename, line, s);
    exit(1);
}


options(argc, argv)
int argc;
char **argv;
{
    char buf[128];

    progname = argv[0];

    while(--argc) {
        argv++;
        if(argv[0][0] != '-')
            break;
        switch(argv[0][1]) {
        case 's':
            sphereprecision = atoi(argv[1]);
            argv++;
            argc--;
            fprintf(stderr, "sphere precision %d\n", sphereprecision);
            break;
        case 'c':
            cylconeprecision = atoi(argv[1]);
            if (cylconeprecision > MAXPANELS)
                cylconeprecision = MAXPANELS;
            argv++;
            argc--;
            fprintf(stderr, "cylcone precision %d\n", cylconeprecision);
            break;
#if( 0 ) 
        case 'r':
            reverse_normals = 1;
            fprintf(stderr, "reverse normals!\n");
            break;
        case 'N':
            ignore_normals = 1;
            fprintf(stderr, "ignoring supplied normals\n");
            break;
#endif
        case 'h':
        case 'H':
            fprintf(stderr,"NFF to TPOLY translator\n");
            fprintf(stderr, "usage: nff2tpoly NFF-file\n");
            exit(0);
            break;
        default:
            sprintf(buf, "unrecognized option: %s", argv[0]);
            fatal(buf);
        }
    }

    if(argc > 1) {
        fprintf(stderr,"NFF to TPOLY translator\n");
        fprintf(stderr, "usage: nff2tpoly NFF-file\n");
        exit(1);
    } else if(argc == 1)
        filename = argv[0];
    else
        filename = "stdin";
}


void
sphere( point p, float r )
{
/*
    translate(p.x, p.y, p.z);
    scale(r, r, r);
*/
    drawsphere( p, r );
}


/* length of a vector */
float
length(point v)
{
    return ( sqrt(v.x*v.x + v.y*v.y + v.z*v.z) );
}


void 
identmatrix(m)
float m[4][4];
{
    int i,j;

    for (i=0;i < 4;i++)
        for (j=0;j < 4;j++)
            m[i][j] = 0.0;
    m[0][0] = m[1][1] = m[2][2] = m[3][3] = 1.0;
}


void
cylcone ( point base , float base_r , point apex , float apex_r )
{
    point len;
    float len0, len1;
    float cosine, sine, incr, theta, dval;
    float points[MAXPANELS*2][3];
    int   i;
    float m[4][4];

    len.x = apex.x - base.x;
    len.y = apex.y - base.y;
    len.z = apex.z - base.z;
    len0  = length(len);

#if( 0 ) 
    if ((len0 = length(len)) < 1.0e-10)
        warn("degenerate cone or cylinder");

    len1 = sqrt(len.x*len.x + len.z*len.z);

    translate(base.x, base.y, base.z);

    identmatrix(m);

    if (len1 == 0.0) {
        cosine = 0.0;
        sine = 1.0;
    } else {
        cosine = len.z / len1;
        sine = len.x / len1;
    }

    /* rot about y */
    m[0][0] = cosine;
    m[0][2] = -sine;
    m[2][0] = sine;
    m[2][2] = cosine;
    multmatrix(m);

    identmatrix(m);

    if (len0 == 0.0) {
        cosine = 0.0;
        sine = 1.0;
    } else {
        cosine = len1 / len0;
        sine = -len.y / len0;
    }

    /* rot about x */
    m[1][1] = cosine;
    m[1][2] = sine;
    m[2][1] = -sine;
    m[2][2] = cosine;
    multmatrix(m);
#endif

    /*
   * Calculate the points on the cylinder
   */
    incr = (2.0 * M_PI)/cylconeprecision;
    theta = 0.0;
    for (i = 0; i < cylconeprecision; i++)
    {
        dval = cos (theta);    /*  x coordinate  */
        points[i][0] = base_r * dval;
        points[i+cylconeprecision][0] = apex_r * dval;
        dval = sin (theta);    /*  y coordinate  */
        points[i][1] = base_r * dval;
        points[i+cylconeprecision][1] = apex_r * dval;
        points[i][2] = 0.0;
        points[i+cylconeprecision][2] = len0;
        theta = theta + incr;
    }

    /*
   * Now display it..
   */
    drawcylinder(cylconeprecision, points, base);
}


/* read in a NFF file */
void
read_nff_file(filename)
char *filename;
{
    FILE *fp;
    char buffer[BUFSIZE];
    int num_vert, tot_vert;
    float kd, ks, shine, t, i_of_r ;
    float v[3], n[3], vert[MAX_VERTS][3];
    point p1, p2;
    float r1, r2;
    point from, at, up;
    float fov, hither;
    float r, g, b;
    int     xres, yres;

    if(strcmp(filename, "stdin") != 0) {
        if ((fp = fopen(filename, "r")) == NULL)
            fatal("failed to open input file for reading");
    } else
        fp = stdin;

    while ( fgets( buffer, BUFSIZE, fp ) ) {
        line++;
        /* polygonal patch */
        if ( sscanf( buffer, "pp %d", &tot_vert ) == 1 ) {
            point *polygon = (point *)malloc( tot_vert * sizeof(point) );
#if( 0 ) 
            if (!ignore_normals) {
#endif
#if( 0 ) 
                bgnpolygon();
#endif
                /* printf( "%d\n", tot_vert ); */
                for ( num_vert = 0 ; num_vert < tot_vert ; ++num_vert ) {
                    fgets( buffer, BUFSIZE, fp ) ;
                    sscanf( buffer, "%g %g %g %g %g %g",
                        &v[0], &v[1], &v[2],
                        &n[0], &n[1], &n[2]);
#if( 0 ) 
                    if (reverse_normals) {
                        vreverse(n);
                    }
                    n3f(n);
#endif
                    polygon[num_vert].x = v[0];
                    polygon[num_vert].y = v[1];
                    polygon[num_vert].z = v[2];
                    /* printf( "%f %f %f\n", v[0], v[1], v[2] ); */
#if( 0 ) 
                    v3f(v);
#endif
                }
                draw_poly( tot_vert, polygon, 1 /* TPOLY */ );
#if( 0 ) 
                endpolygon();
#endif
#if( 0 ) 
            }
            else {
                for ( num_vert = 0 ; num_vert < tot_vert ; ++num_vert ) {
                    fgets( buffer, BUFSIZE, fp ) ;
                    sscanf( buffer, "%g %g %g %g %g %g",
                        &vert[num_vert][0], &vert[num_vert][1], &vert[num_vert][2],
                        &n[0], &n[1], &n[2]); /* ignoring these normals */
                }
#if( 0 ) 
                calc_poly_norm(tot_vert, vert, n);
                if (reverse_normals) {
                    vreverse(n);
                }
#endif
#if( 0 ) 
                bgnpolygon();
#endif
#if( 0 ) 
                n3f(n);
#endif
                /* printf( "%d\n", tot_vert ); */
                for ( num_vert = 0 ; num_vert < tot_vert ; ++num_vert ) {
                    polygon[num_vert].x = vert[num_vert][0];
                    polygon[num_vert].y = vert[num_vert][1];
                    polygon[num_vert].z = vert[num_vert][2];
                    /* printf( "%f %f %f\n", vert[num_vert][0], vert[num_vert][1], vert[num_vert][2] ); */
#if( 0 ) 
                    v3f(&(vert[num_vert][0]));
#endif
                }
                draw_poly( tot_vert, polygon, 1 /* TPOLY */ );
#if( 0 ) 
                endpolygon();
#endif
            }
#endif
        }
        /* polygon */
        else if ( sscanf( buffer, "p %d", &tot_vert ) == 1 ) {
            point *polygon = (point *)malloc( tot_vert * sizeof(point) );

            for ( num_vert = 0 ; num_vert < tot_vert ; ++num_vert ) {
                fgets( buffer, BUFSIZE, fp ) ;
                sscanf( buffer, "%g %g %g",
                    &vert[num_vert][0], &vert[num_vert][1], &vert[num_vert][2]);
            }
#if( 0 ) 
            calc_poly_norm(tot_vert, vert, n);
            if (reverse_normals) {
                vreverse(n);
            }
#endif
#if( 0 ) 
            bgnpolygon();
#endif
#if( 0 ) 
            n3f(n);
#endif
            /* printf( "%d\n", tot_vert ); */
            for ( num_vert = 0 ; num_vert < tot_vert ; ++num_vert ) {
                polygon[num_vert].x = vert[num_vert][0];
                polygon[num_vert].y = vert[num_vert][1];
                polygon[num_vert].z = vert[num_vert][2];
                /* printf( "%f %f %f\n", vert[num_vert][0], vert[num_vert][1], vert[num_vert][2] ); */
#if( 0 ) 
                v3f(&(vert[num_vert][0]));
#endif
            }
            draw_poly( tot_vert, polygon, 1 /* TPOLY */ );
#if( 0 ) 
            endpolygon();
#endif
        }
        /* sphere */
        else if ( sscanf( buffer, "s %g %g %g %g", &p1.x, &p1.y, &p1.z, &r1 ) == 4 )
            /* sphere( p1, r1 ); */
            initsphere(sphereprecision, p1, r1 );
        /* cylinder / cone */
        else if ( buffer[0] == 'c' ) {
            fgets( buffer, BUFSIZE, fp ) ;
            sscanf( buffer, "%g %g %g %g", 
                &p1.x, &p1.y, &p1.z, &r1 ) ;
            fgets( buffer, BUFSIZE, fp ) ;
            sscanf( buffer, "%g %g %g %g", 
                &p2.x, &p2.y, &p2.z, &r2 ) ;
            cylcone( p1, r1, p2, r2 ) ;
        }
        /* fill color */
        else if ( buffer[0] == 'f' ) ;
        /* light */
        else if ( buffer[0] == 'l' ) ;
        /* view */
        else if ( buffer[0] == 'v' ) ;
        /* background color */
        else if ( buffer[0] == 'b' ) ;
        /* comment! */
        else if ( buffer[0] == '#' ) ;
        /* error */
        else ;
    }
    fclose(fp);
}


/* subtract two vectors to form a third */
void 
subv ( point *o , point *a , point *b )
{
    o->x = a->x - b->x;
    o->y = a->y - b->y;
    o->z = a->z - b->z;
}


/* normalize a vector */
point
*normalize ( point *a )
{
    static point    r;
    double          mag;

    r = *a;
    mag = r.x * r.x + r.y * r.y + r.z * r.z;
    if (mag != 0.0) {
        mag = 1.0 / sqrt(mag);
        r.x *= mag;
        r.y *= mag;
        r.z *= mag;
    }
    return &r;
}


/* calculate cross product of 2 vectors */
void 
crossp ( point *o , point *a , point *b )
{
    o->x = (a->y * b->z) - (a->z * b->y);
    o->y = (a->z * b->x) - (a->x * b->z);
    o->z = (a->x * b->y) - (a->y * b->x);
    *o = *normalize(o);
}


/* Return the midpoint on the line between two points */
point          *
    midpoint(a, b)
    point          *a, *b;
{
    static point    r;

    r.x = (a->x + b->x) * 0.5;
    r.y = (a->y + b->y) * 0.5;
    r.z = (a->z + b->z) * 0.5;

    return &r;
}


/* normalize an array which represents a vector */
float 
anormalize ( float *v )
{
    double          d;

    d = sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2]);
    if (fabs(d) > 1.0e-6) {
        v[0] /= d;
        v[1] /= d;
        v[2] /= d;
    }
    return (d);
}


vreverse(float v[3])
    {
    v[0] = -1.0 * v[0];
    v[1] = -1.0 * v[1];
    v[2] = -1.0 * v[2];
}


calc_poly_norm(int nverts, float v[][3], float n[3])
{
    int i, j;

    /* Martin Newell's method:  Newman & Sproull page 499
   ** More robust than the 3 point method.  
   */
    n[0] = n[1] = n[2] = 0.0;

    for (j=1,i=0; i < nverts; j++,i++) {
        if(j == nverts)
            j = 0;
        n[0] += (v[i][1] - v[j][1]) * (v[i][2] + v[j][2]);
        n[1] += (v[i][2] - v[j][2]) * (v[i][0] + v[j][0]);
        n[2] += (v[i][0] - v[j][0]) * (v[i][1] + v[j][1]);
    }
    anormalize(n);
}


void 
make_triangle(t)
triangle *t;
{
    int   i;
    float v[3];

#if( 0 ) 
    bgnpolygon();
#endif
    printf( "3\n" );
    for (i = 0; i < 3; i++) {
        v[0] = t->pt[i].x;
        v[1] = t->pt[i].y;
        v[2] = t->pt[i].z;
#if( 0 ) 
        n3f(v);
#endif
        printf( "%f %f %f\n", v[0], v[1], v[2] );
#if( 0 ) 
        v3f(v);
#endif
    }
#if( 0 ) 
    endpolygon();
#endif
}


void 
make_object(obj, level)
object *obj;
int     level;
{
    int i;

    /* Spit out coordinates for each triangle */
    for (i = 0; i < obj->npoly; i++)
        make_triangle(&obj->poly[i]);
}


/*
 * generate a triangle mesh approximating a sphere by recursive
 * subdivision. First approximation is an octahedron; each level of
 * refinement increases the number of triangles by a factor of 4. Level 3
 * (128 triangles) is a good tradeoff if gouraud shading is used to render
 * the database.
 *
 * based on program sphere, Jon Leech (leech@cs.unc.edu)
 */
initsphere(int maxlevel, point p, float r)
{
    object *old, *new;
    int i, level;

    old = &oct;
    /* Subdivide each starting triangle (maxlevel - 1) times */
    for (level = 1; level < maxlevel; level++) {
        /* Allocate a new object */
        new = (object *) malloc(sizeof(object));
        if (new == NULL) {
            fprintf(stderr, "initsphere: Out of memory on subdivision level %d\n",
                level);
            exit(1);
        }
        new->npoly = old->npoly * 4;

        /*
     * Allocate 4* the number of points in the current
     * approximation
     */
        new->poly = (triangle *) malloc(new->npoly * sizeof(triangle));
        if (new->poly == NULL) {
            fprintf(stderr, "initsphere: Out of memory on subdivision level %d\n",
                level);
            exit(1);
        }
        /*
     * Subdivide each triangle in the old approximation and
     * normalize the new points thus generated to lie on the
     * surface of the unit sphere. Each input triangle with
     * vertices labelled [0,1,2] as shown below will be turned
     * into four new triangles:
     * 
     * Make new points a = (0+2)/2 b = (0+1)/2 c = (1+2)/2 1 /\
     * normalize a, b, c /  \ b/____\ c        Construct new
     * triangles /\    /\            [0,b,a] /    \  /  \
     * [b,1,c] /____\/____\        [a,b,c] 0      a    2
     * [a,c,2]
     */
        for (i = 0; i < old->npoly; i++) {
            triangle
            * oldt = &old->poly[i],
            *newt = &new->poly[i * 4];
            point           a, b, c;

            a = *normalize(midpoint(&oldt->pt[0], &oldt->pt[2]));
            b = *normalize(midpoint(&oldt->pt[0], &oldt->pt[1]));
            c = *normalize(midpoint(&oldt->pt[1], &oldt->pt[2]));

            newt->pt[0] = oldt->pt[0];
            newt->pt[1] = b;
            newt->pt[2] = a;
            newt++;

            newt->pt[0] = b;
            newt->pt[1] = oldt->pt[1];
            newt->pt[2] = c;
            newt++;

            newt->pt[0] = a;
            newt->pt[1] = b;
            newt->pt[2] = c;
            newt++;

            newt->pt[0] = a;
            newt->pt[1] = c;
            newt->pt[2] = oldt->pt[2];
        }

        if (level > 1) {
            free(old->poly);
            free(old);
        }
        /* Continue subdividing new triangles */
        old = new;
    }
    make_object(old, maxlevel);
}


/*    drawcylinder (sides, points)
 *    sides are the number of panels around the circular perimeter.
 *    points is the array of the coordinates of the vertices.
 *    This subroutine draws a polygonal cylinder.
 *
 *    The cylinder is circled by rectangles.  The final rectangle is
 *      drawn outside the loop because it connects the first and last
 *      points.
 */
drawcylinder (sides, points, base)
int sides;
float points[][3];
point base;
{
    int i;
    float n[3];
    point polygon[4];

    /* draw the side facets of the cylinder */
    for (i = 0; i < sides - 1; i = i + 1)
    {
        /* printf( "4\n" ); */
#if( 0 ) 
        bgnpolygon();
#endif
#if( 0 ) 
        n[0] = points[i][0];
        n[1] = points[i][1];
        n[2] = 0.0;
        anormalize(n);
        n3f( n );
#endif
        polygon[0].x = points[i][0];
        polygon[0].y = points[i][1];
        polygon[0].z = points[i][2];
        /* printf( "%f %f %f\n", points[i][0], points[i][1], points[i][2] ); */
#if( 0 ) 
        v3f( points[i] );
#endif
#if( 0 ) 
        n[0] = points[sides + i][0];
        n[1] = points[sides + i][1];
        anormalize(n);
        n3f( n );
#endif
        polygon[1].x = points[sides+i][0];
        polygon[1].y = points[sides+i][1];
        polygon[1].z = points[sides+i][2];
        /* printf( "%f %f %f\n", points[sides+i][0], points[sides+i][1], points[sides+i][2] ); */
#if( 0 ) 
        v3f( points[sides+i] );
#endif
#if( 0 ) 
        n[0] = points[sides + i + 1][0];
        n[1] = points[sides + i + 1][1];
        anormalize(n);
        n3f( n );
#endif
        polygon[2].x = points[sides+i+1][0];
        polygon[2].y = points[sides+i+1][1];
        polygon[2].z = points[sides+i+1][2];
        /* printf( "%f %f %f\n", points[sides+i+1][0], points[sides+i+1][1], points[sides+i+1][2] ); */
#if( 0 ) 
        v3f( points[sides+i+1] );
#endif
#if( 0 ) 
        n[0] = points[i + 1][0];
        n[1] = points[i + 1][1];
        anormalize(n);
        n3f( n );
#endif
        polygon[3].x = points[i+1][0];
        polygon[3].y = points[i+1][1];
        polygon[3].z = points[i+1][2];
        /* printf( "%f %f %f\n", points[i+1][0], points[i+1][1], points[i+1][2] ); */
#if( 0 ) 
        v3f( points[i+1] );
#endif
#if( 0 ) 
        endpolygon();
#endif
        draw_poly( 4, polygon, 1 /* TPOLY */ );
    }

    /*  draw final side facet which connects first and last points    */
    /* printf( "4\n" ); */
#if( 0 ) 
    bgnpolygon();
#endif
#if( 0 ) 
    n[0] = points[sides-1][0];
    n[1] = points[sides-1][1];
    anormalize(n);
    n3f( n );
#endif
    polygon[0].x = points[sides-1][0];
    polygon[0].y = points[sides-1][1];
    polygon[0].z = points[sides-1][2];
    /* printf( "%f %f %f\n", points[sides-1][0], points[sides-1][1], points[sides-1][2] ); */
#if( 0 ) 
    v3f( points[sides-1] );
#endif
#if( 0 ) 
    n[0] = points[2*sides-1][0];
    n[1] = points[2*sides-1][1];
    anormalize(n);
    n3f( n );
#endif
    polygon[1].x = points[2*sides-1][0];
    polygon[1].y = points[2*sides-1][1];
    polygon[1].z = points[2*sides-1][2];
    /* printf( "%f %f %f\n", points[2*sides-1][0], points[2*sides-1][1], points[2*sides-1][2] ); */
#if( 0 ) 
    v3f( points[2*sides-1] );
#endif
#if( 0 ) 
    n[0] = points[sides][0];
    n[1] = points[sides][1];
    anormalize(n);
    n3f( n );
#endif
    polygon[2].x = points[sides][0];
    polygon[2].y = points[sides][1];
    polygon[2].z = points[sides][2];
    /* printf( "%f %f %f\n", points[sides][0], points[sides][1], points[sides][2] ); */
#if( 0 ) 
    v3f( points[sides] );
#endif
#if( 0 ) 
    n[0] = points[0][0];
    n[1] = points[0][1];
    anormalize(n);
    n3f( n );
#endif
    polygon[3].x = points[0][0];
    polygon[3].y = points[0][1];
    polygon[3].z = points[0][2];
    /* printf( "%f %f %f\n", points[0][0], points[0][1], points[0][2] ); */
#if( 0 ) 
    v3f( points[0] );
#endif
#if( 0 ) 
    endpolygon();
#endif
    draw_poly( 4, polygon, 1 /* TPOLY */ );
}


drawsphere ( point p, float r )
{
    if (!made_sphere) {
        initsphere(sphereprecision, p, r );
        made_sphere = 1;
    }
}


main(argc, argv)
int    argc;
char    **argv;
{
    options(argc, argv);
    read_nff_file(filename);
}
