
#include <X11/Xlib.h>
#include "color.h"

#define    MIN(a,b)    ((a)<(b)?(a):(b))
#define    MAX(a,b)    ((a)>(b)?(a):(b))
#define    ABS(x)      ((x)<0?-(x):(x))

/*
 * Mix two RGBs, with scale factors alpha and beta, in RGB space.
 */

RGB
MixRGB(r, alpha, s, beta)
RGB    r, s;
float    alpha, beta;
{
    RGB    t;

    t.r = MAX(0, MIN(MAX_INTENSITY, (int)(alpha*(r.r) + beta*(s.r))));
    t.g = MAX(0, MIN(MAX_INTENSITY, (int)(alpha*(r.g) + beta*(s.g))));
    t.b = MAX(0, MIN(MAX_INTENSITY, (int)(alpha*(r.b) + beta*(s.b))));
    return t;
}

/*
 * Mix two RGBs with scale factors alpha and beta, in HSV space.
 */

RGB
MixHSV(r, alpha, s, beta)
RGB    r, s;
float    alpha, beta;
{
    HSV    rr, ss, tt;

    rr = RGBToHSV(r);
    ss = RGBToHSV(s);
    tt.h = alpha*rr.h + beta*ss.h;
    if (ABS(rr.h - ss.h) > 0.5) {
        tt.h = tt.h + 0.5;
        if (tt.h >= 1.0)
            tt.h = tt.h - 1.0;
    }
    tt.s = alpha*rr.s + beta*ss.s;
    tt.v = alpha*rr.v + beta*ss.v;
    return HSVToRGB(tt);
}

/*
 * Convert an HLS to an RGB.
 */

float Value( n1, n2, hue )
float n1, n2, hue;
{
    if( hue > 1. ) hue -= 1.;
    if( hue < 0. ) hue += 1.;
    if( hue < 0.1667 )
        return( n1 + ( n2 - n1 ) * hue / 0.1667 );
    else if( hue < 0.50 )
        return( n2 );
    else if( hue < 0.6667 )
        return( n1 + ( n2 - n1 ) * (0.6667 - hue) / 0.1667 );
    else
        return( n1 );
}

RGB
HLSToRGB(hls)
HLS    hls;
{
    RGB    rgb;
    float m1, m2;

    if( hls.l <= 0.5 )
        m2 = hls.l * (1 + hls.s);
    else
        m2 = hls.l + hls.s - hls.l * hls.s;
    m1 = 2 * hls.l - m2;
    if (hls.s == 0.0)
        rgb.r = rgb.g = rgb.b = (int)(0.5 + hls.l * MAX_INTENSITY);
    else {
        rgb.r = (int)(0.5 + Value( m1, m2, hls.h + 0.333 ) * MAX_INTENSITY);
        rgb.g = (int)(0.5 + Value( m1, m2, hls.h )         * MAX_INTENSITY);
        rgb.b = (int)(0.5 + Value( m1, m2, hls.h - 0.333 ) * MAX_INTENSITY);
    }

    return rgb;
}

HLS
RGBToHLS(rgb)
RGB rgb;
{
    HLS   hls;
    float mx, mn;
    float rr, gg, bb;
    float rc, gc, bc;

    rr = (float) rgb.r / (float) MAX_INTENSITY;
    gg = (float) rgb.g / (float) MAX_INTENSITY;
    bb = (float) rgb.b / (float) MAX_INTENSITY;

    mx = MAX(MAX(rr, gg), bb);
    mn = MIN(MIN(rr, gg), bb);
    hls.l = (mx + mn) / 2.;
    if( mx == mn )
        hls.s = 0.0;
    else {
        if( hls.l <= 0.5 ) hls.s = (mx - mn) / (mx + mn);
        else               hls.s = (mx - mn) / (2. - mx - mn);
        rc = (mx - rr) / (mx - mn);
        gc = (mx - gg) / (mx - mn);
        bc = (mx - bb) / (mx - mn);
             if( rr == mx ) hls.h = bc - gc;
        else if( gg == mx ) hls.h = 2. + rc - bc;
        else if( bb == mx ) hls.h = 4. + gc - rc;
        hls.h *= 0.166666;
        if( hls.h < 0. ) hls.h += 1.;
    }

    return hls;
}

/*
 * Convert an HSV to an RGB.
 */

RGB
HSVToRGB(hsv)
HSV    hsv;
{
    RGB    rgb;
    float    p, q, t, f;
    int    i;

    if (hsv.s == 0.0)
        rgb = PctToRGB(hsv.v, hsv.v, hsv.v);
    else {
        if (hsv.s > 1.0)
            hsv.s = 1.0;
        if (hsv.s < 0.0)
            hsv.s = 0.0;
        if (hsv.v > 1.0)
            hsv.v = 1.0;
        if (hsv.v < 0.0)
            hsv.v = 0.0;
        if (hsv.h >= 1.0)
            hsv.h = 0.0;

        hsv.h = 6.0 * hsv.h;
        i = (int) hsv.h;
        f = hsv.h - (float) i;
        p = hsv.v * (1.0 - hsv.s);
        q = hsv.v * (1.0 - (hsv.s * f));
        t = hsv.v * (1.0 - (hsv.s * (1.0 - f)));

        switch(i) {
        case 0:    
            rgb = PctToRGB(hsv.v, t, p); 
            break;
        case 1:    
            rgb = PctToRGB(q, hsv.v, p); 
            break;
        case 2:    
            rgb = PctToRGB(p, hsv.v, t); 
            break;
        case 3:    
            rgb = PctToRGB(p, q, hsv.v); 
            break;
        case 4:    
            rgb = PctToRGB(t, p, hsv.v); 
            break;
        case 5:    
            rgb = PctToRGB(hsv.v, p, q); 
            break;
        }
    }
    return rgb;
}

/*
 * Convert an RGB to HSV.
 */

HSV
RGBToHSV(rgb)
RGB    rgb;
{
    HSV    hsv;
    float    rr, gg, bb;
    float    min, max;
    float    rc, gc, bc;

    rr = (float) rgb.r / (float) MAX_INTENSITY;
    gg = (float) rgb.g / (float) MAX_INTENSITY;
    bb = (float) rgb.b / (float) MAX_INTENSITY;

    max = MAX(MAX(rr, gg), bb);
    min = MIN(MIN(rr, gg), bb);
    hsv.v = max;
    if (max == 0.0)
        hsv.s = 0.0;
    else
        hsv.s = (max - min) / max;
    if (hsv.s == 0.0)
        hsv.h = 0.0;
    else {
        rc = (max - rr) / (max - min);
        gc = (max - gg) / (max - min);
        bc = (max - bb) / (max - min);
        if (rr == max)
            hsv.h = bc - gc;
        else if (gg == max)
            hsv.h = 2.0 + rc - bc;
        else if (bb = max)
            hsv.h = 4.0 + gc - rc;

        if (hsv.h < 0.0)
            hsv.h += 6.0;
        hsv.h = hsv.h / 6.0;
    }

    return hsv;
}

/*
 * Intensity percentages to RGB.
 */

RGB
PctToRGB(rr, gg, bb)
float    rr, gg, bb;
{
    RGB    rgb;

    if (rr > 1.0)
        rr = 1.0;
    if (gg > 1.0)
        gg = 1.0;
    if (bb > 1.0)
        bb = 1.0;

    rgb.r = (int)(0.5 + rr * MAX_INTENSITY);
    rgb.g = (int)(0.5 + gg * MAX_INTENSITY);
    rgb.b = (int)(0.5 + bb * MAX_INTENSITY);
    return rgb;
}

/*
 * Intensity percentages to HSV.
 */

HSV
PctToHSV(hh, ss, vv)
float    hh, ss, vv;
{
    HSV    hsv;

    if (hh > 1.0)
        hh = 1.0;
    if (ss > 1.0)
        ss = 1.0;
    if (vv > 1.0)
        vv = 1.0;

    hsv.h = hh;
    hsv.s = ss;
    hsv.v = vv;
    return hsv;
}

/*
 * The Manhattan distance between two colors, between 0.0 and 3.0.

float
RGBDist(r, s)
RGB    r, s;
{
    return (
        (float)ABS(r.r - s.r) +
        (float)ABS(r.g - s.g) +
        (float)ABS(r.b - s.b)) / (float)MAX_INTENSITY;
}
 */

/*
 * Load an XColor with an RGB.
 */

RGBToXColor(r, x)
RGB    r;
XColor    *x;
{
    x->red = r.r;
    x->green = r.g;
    x->blue = r.b;
    x->flags = DoRed | DoGreen | DoBlue;
}

/*
 * Convert a CMY to RGB.
 */

RGB
CMYToRGB(cmy)
CMY    cmy;

{
    RGB    rgb;

    rgb.r = MAX_INTENSITY - cmy.c;
    rgb.g = MAX_INTENSITY - cmy.m;
    rgb.b = MAX_INTENSITY - cmy.y;
    return rgb;
}

/*
 * Convert an RGB to CMY.
 */

CMY
RGBToCMY(rgb)
RGB    rgb;

{
    CMY    cmy;

    cmy.c = MAX_INTENSITY - rgb.r;
    cmy.m = MAX_INTENSITY - rgb.g;
    cmy.y = MAX_INTENSITY - rgb.b;
    return cmy;
}

/*
 * Convert an CIE to RGB
 */

/*
 *                   | 2.750 -1.118  0.138|
 * [R G B] = [X Y Z] |-1.149  2.026 -0.333|
 *                   |-0.426  0.033  1.104|
 */

RGB
CIEToRGB(cie)
CIE cie;
{
    RGB rgb;

    rgb.r = ( cie.X * 2.750 - cie.Y * 1.149 - cie.Z * 0.426) * 256;
    rgb.g = (-cie.X * 1.118 + cie.Y * 2.026 + cie.Z * 0.033) * 256;
    rgb.b = ( cie.X * 0.138 - cie.Y * 0.333 + cie.Z * 1.104) * 256;
/*  rgb.b = (int)(0.5 + ( cie.X * 0.138 - cie.Y * 0.333 + cie.Z * 1.105) / 255. * MAX_INTENSITY); */
    return rgb;
}

RGBf
CIEToRGBs(cie)
CIE cie;
{
    RGBf rgb;

    rgb.r =  cie.X * 2.750 - cie.Y * 1.149 - cie.Z * 0.426;
    rgb.g = -cie.X * 1.118 + cie.Y * 2.026 + cie.Z * 0.033;
    rgb.b =  cie.X * 0.138 - cie.Y * 0.333 + cie.Z * 1.104;
    return rgb;
}

/*
 *                   | 0.478  0.263  0.020|
 * [X Y Z] = [R G B] | 0.299  0.655  0.160|
 *                   | 0.175  0.081  0.908|
 */

CIE
RGBToCIE(rgb)
RGB rgb;
{
    CIE cie;
    float rr, gg, bb;

    rr = (float) rgb.r / (float) MAX_INTENSITY;
    gg = (float) rgb.g / (float) MAX_INTENSITY;
    bb = (float) rgb.b / (float) MAX_INTENSITY;

    cie.X = rr * 0.478 + gg * 0.299 + bb * 0.175;
    cie.Y = rr * 0.263 + gg * 0.655 + bb * 0.081;
    cie.Z = rr * 0.020 + gg * 0.160 + bb * 0.908;
    return cie;
}
